function [] = Nano_Analysis(file_name, dia_particle)
% This function is to detect all the nanoparticles in a image.
% 'file_name' is the name of the image file
% 'dia_particle' is the estimated value of the diameter of nano-particles; the default set is 20
% Example: Nano_Analysis('CAS_F3-2_7.tif', 50) or Nano_Analysis('CAS_F10_8.tif', 20) 

addpath(genpath('Functions'));

% Set some parameters
b_Bandpass = 1; % whether to use the bandpass filter
b_Gaussian = 1; % whether to use the Gaussian filter
b_reshape = 0; % whether to reshape the image
b_intermediary = 0; % whether to save those intermediary results

% Test time
tic;

% Calculate the minimum value of area
area_min = (dia_particle / 2)^2 / 2;

% Set default value of 'dia_particle'
if nargin < 2
    dia_particle = 20;
elseif nargin < 1
    error(message('Not enough input'));
end

close all;
% Read the image
[~,image_name,~] = fileparts(file_name);
Raw_image = imread(file_name);

if ~ismatrix(Raw_image)
    Raw_image = rgb2gray(Raw_image);
end
Raw_image = double(Raw_image);
[n_height, n_width] = size(Raw_image);

if b_reshape == 1
    if max(n_height, n_width) > 1024
        eff_scale = 1024.0 / double(max(n_height, n_width));
        Raw_image = imresize(Raw_image, eff_scale);
        
        [n_height, n_width] = size(Raw_image);
        dia_particle = round(eff_scale * dia_particle);
    end
end

% Step 1-1: FFT-Bandpass filtering
fprintf('1 Preprocessing...\n');
if b_Bandpass == 1
    fprintf('1-1 FFT-Bandpass filtering...\n');
    
    % Try to remove edge effect
    % Add a edge
    width_edge = 100;
    Temp_image = zeros(n_height + 2 * width_edge, n_width + 2 * width_edge);
    Temp_image(1 + width_edge : n_height + width_edge, 1 + width_edge : n_width + width_edge) = Raw_image;
    Temp_image(1 + width_edge : n_height + width_edge, 1 : width_edge) = repmat(Temp_image(1 + width_edge : n_height + width_edge, 1 + width_edge), 1, width_edge);
    Temp_image(1 + width_edge : n_height + width_edge, n_width + width_edge + 1 : n_width + 2 * width_edge) = repmat(Temp_image(1 + width_edge : n_height + width_edge, n_width + width_edge), 1, width_edge);
    Temp_image(1: width_edge, 1 + width_edge : n_width + width_edge) = repmat(Temp_image(1 + width_edge, 1 + width_edge : n_width + width_edge), width_edge, 1);
    Temp_image(n_height + width_edge + 1 : n_height + 2 * width_edge, 1 + width_edge : n_width + width_edge) = repmat(Temp_image(n_height + width_edge, 1 + width_edge : n_width + width_edge), width_edge, 1);
    Temp_image(1: width_edge, 1:width_edge) = Temp_image(width_edge + 1, width_edge + 1);
    Temp_image(1: width_edge, n_width + width_edge + 1 : n_width + 2 * width_edge) = Temp_image(width_edge + 1, width_edge + n_height);
    Temp_image(n_height + width_edge + 1 : n_height + 2 * width_edge, 1 : width_edge) = Temp_image(width_edge + n_height, width_edge + 1);
    Temp_image(n_height + width_edge + 1 : n_height + 2 * width_edge, n_width + width_edge + 1 : n_width + 2 * width_edge) = Temp_image(width_edge + n_height,width_edge + n_height);
    
    Fre_image = fft2(Temp_image);
    Fre_image = fftshift(Fre_image);
    
    [n_height, n_width] = size(Temp_image);
    half_height = fix(n_height / 2) + 1;
    half_width = fix(n_width / 2) + 1;
    Mask_image = ones(n_height, n_width);
    
    % low_fre = 15;
    low_fre = 256 / (double(dia_particle)) / 1024.0 * double(min(n_width, n_height));
    high_fre = 128;
    
    % Bandpass filtering
    for i = 1:n_height
        for j = 1:n_width
            dis = sqrt((i - half_height)^2 + (j - half_width)^2);
            if dis == 0
                Mask_image(i, j) = 1.1;
            else
                eff_1 = 1.414 / (1 + 0.414 * (dis / high_fre) ^ 4);
                eff_2 = 1.414 / (1 + 0.414 * (low_fre / dis) ^ 4);
                Mask_image(i, j) = eff_1 * eff_2;
            end
        end
    end
    
    Fre_image_filtered = Fre_image .* Mask_image;
    
    Fre_image_filtered = fftshift(Fre_image_filtered);
    Image_filtered = ifft2(Fre_image_filtered);
    
    [n_height, n_width] = size(Raw_image);
    Image_filtered = real(Image_filtered(width_edge + 1 : width_edge + n_height, width_edge + 1 : width_edge + n_width));
    Image_filtered = uint8(Image_filtered);
else
    Image_filtered = uint8(Raw_image);
end

if b_intermediary == 1
    imwrite(Image_filtered, [image_name, '_Bandpass.jpg'], 'jpg');
end

% Step 1-2: Gaussian filter 
if b_Gaussian == 1
    fprintf('1-2 Gaussian filtering...\n');
    
    hsize = 21;
    sigma = 2 * dia_particle / 20;
    h1 = fspecial('gaussian', hsize, sigma);
    Image_filtered = imfilter(Image_filtered, h1, 'replicate');
end
if b_intermediary == 1
    imwrite(Image_filtered, [image_name, '_Gaussian.jpg'], 'jpg');
end
% return;

% A_path: intensity-based particle detection
A_path = 1;
if A_path == 1
    fprintf('2 Processing image via the intensity based method...\n');
    
    % Step 2-1: Segmenting foreground/background via k-means
    
    % Local segmentation method (K-means)
    fprintf('2-1 Segmenting foreground/background via k-means...\n');
    gray_data = zeros(n_height * n_width, 3);
    gray_data(:, 1) = double(reshape(repmat(1:n_width, n_height, 1), n_height * n_width, 1)) / double(n_width);
    gray_data(:, 2) = double(reshape(repmat(1:n_height, n_width, 1)', n_height * n_width, 1)) / double(n_height);
    gray_data(:, 3) = double(reshape(Image_filtered, n_height * n_width, 1)) / 255.0;
    gray_level = double(reshape(Image_filtered, n_height * n_width, 1));
    
    weight_space = 0.1 * 60.0 / double(dia_particle) / 1024.0 * double(min(n_width, n_height));
    if weight_space > 0.1
        weight_space = 0.1;
    end
    gray_data(:, 1) = gray_data(:, 1) .* weight_space ;
    gray_data(:, 2) = gray_data(:, 2) .* weight_space;
    
    Inital_Matrix = [0.5 0.5 1.0; 0.5 0.5 0.0];
    IDX_data = kmeans(gray_data, 2, 'start', Inital_Matrix);
    gray_level1 = mean(gray_level(IDX_data == 1));
    gray_level2 = mean(gray_level(IDX_data == 2));
    IDX_Image = reshape(IDX_data, n_height, n_width);
    Bi_image = zeros(n_height, n_width);
    if gray_level1 > gray_level2
        Bi_image(IDX_Image == 1) = 1;
    else
        Bi_image(IDX_Image == 2) = 1;
    end
    if b_intermediary == 1
        imwrite(Bi_image, [image_name, '_Binary_Local.jpg'], 'jpg');
    end
    
    % Step 2-2: Watershed to segment the overlapped particles
    fprintf('2-2 Segmenting overlapped particles via watershed method...\n');
    Bi_image = (Bi_image > 0);
    Re_image_Cov = seg_particle_Convex_Park(Bi_image, Raw_image, dia_particle);
    if b_intermediary == 1
        imwrite(Re_image_Cov, [image_name, '_Seg_Cov.jpg'], 'jpg');
    end
    
    % return;
end

% B_path: gradient-based particle detection
B_path = 1;
if B_path == 1
    fprintf('3 Processing image via the gradient based method...\n');
    
    % Step 3-1: Active contour to find the edge
    % Find the initial mask
    fprintf('3-1 Segmenting foreground/background via active contour...\n');
    % fprintf('Shrinking from a larger mask...\n');
    off_set = 15;
    level = graythresh(Image_filtered) + 1.0 * double(off_set) / 255.0;
    mask = imbinarize(Image_filtered,level);
    if b_intermediary == 1
        imwrite(mask, [image_name, '_Mask_1.jpg'], 'jpg');
    end
    
    % tic
    Bi_image_1 = Active_Contour(Image_filtered, mask, dia_particle);
    Bi_image_1 = (Bi_image_1 > 0);
    % toc
    if b_intermediary == 1
        imwrite(Bi_image_1, [image_name, '_Contour_1.jpg'], 'jpg');
    end
    
    % fprintf('Expanding from a smaller mask...\n');
    level = graythresh(Image_filtered) - 1.5 * double(off_set) / 255.0;
    mask = imbinarize(Image_filtered, level);
    if b_intermediary == 1
        imwrite(mask, [image_name, '_Mask_2.jpg'], 'jpg');
    end
    
    % tic
    Bi_image_2 = Active_Contour(Image_filtered, mask, dia_particle);
    Bi_image_2 = (Bi_image_2 > 0);
    % toc
    if b_intermediary == 1
        imwrite(Bi_image_2, [image_name, '_Contour_2.jpg'], 'jpg');
    end
    
    % Use two direction to find a better result
    % fprintf('Synthesizing the two contours...\n');
    Bi_image_1 = imcomplement(Bi_image_1);
    Bi_image_2 = imcomplement(Bi_image_2);
    SE = strel('disk', round(double(dia_particle) / 10.0));
    Bi_image_2 = imdilate(Bi_image_2, SE);
    Bi_image_1 = Bi_image_1 .* Bi_image_2;
    
    Bi_image = imcomplement(Bi_image_1);
    if b_intermediary == 1
        imwrite(Bi_image, [image_name, '_Contour.jpg'], 'jpg');
    end
    % return;
    
    % Step 3-2: Iterative voting
    fprintf('3-2 Segmenting overlapped particles via iterative voting...\n');
    %tic;
    Voted_center = Iterative_Voting(Image_filtered, dia_particle);
    %toc;
    
    Mark_image = zeros(n_height, n_width);
    num_Local_max = sum(sum(Voted_center));
    [y_Local_max, x_Local_max] = find(Voted_center > 0);
    centroids = [x_Local_max y_Local_max];
    
    % Draw the centroids
    if b_intermediary == 1
        cross_width = 1;
        cross_height = 8;
        for k = 1:num_Local_max
            
            y = round(centroids(k, 2));
            x = round(centroids(k, 1));
            
            i_begin = max(1, y - cross_height);
            j_begin = max(1, x - cross_height);
            i_end = min(n_height, y + cross_height);
            j_end = min(n_width, x + cross_height);
            
            for i = i_begin:i_end
                for j = j_begin:j_end
                    
                    now_width = min(abs(i - y), abs(j - x));
                    now_height = max(abs(i - y), abs(j - x));
                    if (now_width <= cross_width) && (now_height <= cross_height)
                        Mark_image(i, j) = 1;
                    end
                    
                end
            end
        end
        
        Result_image = zeros(n_height, n_width, 3);
        Result_image(:, :, 1) = (Mark_image  == 0) .* Raw_image + (Mark_image  == 1) .* 255;
        Result_image(:, :, 2) = (Mark_image  == 0) .* Raw_image;
        Result_image(:, :, 3) = (Mark_image  == 0) .* Raw_image;
        Result_image = uint8(Result_image);
        imwrite(Result_image, [image_name, '_IV.jpg'], 'jpg');
    end
    % return
    
    % Segmentation with the direction of the contours
    Re_image_IV = seg_particle_IV(Bi_image, Voted_center, dia_particle);
    Re_image_IV2 = imcomplement(Re_image_IV);
    Re_image_IV2  = bwareaopen(Re_image_IV2 , round(area_min), 4);
    Re_image_IV = imcomplement(Re_image_IV2);
    
    if b_intermediary == 1
        imwrite(Re_image_IV, [image_name, '_Seg_IV.jpg'], 'jpg');
    end
    
    % return;
end

% Step 4: Fit the ellipses for each particle
fprintf('4 Postprocessing: fitting each particle by a elliptical model...\n');

% From the results of the intensity-based method
if b_intermediary == 1
    Re_image_Cov = imread([image_name, '_Seg_Cov.jpg'], 'jpg');
    Re_image_Cov = logical(Re_image_Cov > 127);
end

[Ell_Para_Cov, Ell_Contour_Cov, Ell_Rel_Cov] = Ellipse_Fit(Re_image_Cov, dia_particle);

% Output result
if b_intermediary == 1
    Mark_image = zeros(n_height, n_width);
    line_width = 1;
    for k = 1:length(Ell_Contour_Cov)
        boundary = Ell_Contour_Cov{k};
        for l = 1 : length(boundary)
            
            y = boundary(l,1);
            x = boundary(l,2);
            
            i_begin = max(1, y - line_width);
            j_begin = max(1, x - line_width);
            i_end = min(n_height, y + line_width);
            j_end = min(n_width, x + line_width);
            for i = i_begin:i_end
                for j = j_begin:j_end
                    
                    now_width = sqrt((i - y)^2 + (j - x)^2);
                    
                    if now_width <= line_width
                        Mark_image(i, j) = 2 + Ell_Rel_Cov(k, 1);
                    end
                end
            end
        end
    end
    cross_width = 1;
    cross_height = 8;
    for k = 1:length(Ell_Contour_Cov)
        
        y = round(Ell_Para_Cov(k).y);
        x = round(Ell_Para_Cov(k).x);
        
        i_begin = max(1, y - cross_height);
        j_begin = max(1, x - cross_height);
        i_end = min(n_height, y + cross_height);
        j_end = min(n_width, x + cross_height);
        
        for i = i_begin:i_end
            for j = j_begin:j_end
                
                now_width = min(abs(i - y), abs(j - x));
                now_height = max(abs(i - y), abs(j - x));
                if (now_width <= cross_width) && (now_height <= cross_height)
                    Mark_image(i, j) = 1;
                end
            end
        end
    end
    
    Result_image = zeros(n_height, n_width, 3);
    Result_image(:, :, 1) = (Mark_image == 0) .* Raw_image + (Mark_image == 1) .* 255;
    Result_image(:, :, 2) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .*  round(255 .* (Mark_image - 2));
    Result_image(:, :, 3) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .* (128 + round(127 .* (3 - Mark_image)));
    Result_image = uint8(Result_image);
    imwrite(Result_image, [image_name, '_Fit_Cov.jpg'], 'jpg');
end

% From the results of the gradient-based method
if b_intermediary == 1
    Re_image_IV = imread([image_name, '_Seg_IV.jpg'], 'jpg');
    Re_image_IV = logical(Re_image_IV > 127);
end
[Ell_Para_IV, Ell_Contour_IV, Ell_Rel_IV] = Ellipse_Fit(Re_image_IV, dia_particle);

% Output result
if b_intermediary == 1
    Mark_image = zeros(n_height, n_width);
    line_width = 1;
    for k = 1:length(Ell_Contour_IV)
        boundary = Ell_Contour_IV{k};
        for l = 1 : length(boundary)
            
            y = boundary(l,1);
            x = boundary(l,2);
            
            i_begin = max(1, y - line_width);
            j_begin = max(1, x - line_width);
            i_end = min(n_height, y + line_width);
            j_end = min(n_width, x + line_width);
            for i = i_begin:i_end
                for j = j_begin:j_end
                    
                    now_width = sqrt((i - y)^2 + (j - x)^2);
                    
                    if now_width <= line_width
                        Mark_image(i, j) = 2 + Ell_Rel_IV(k, 1);
                    end
                end
            end
        end
    end
    cross_width = 1;
    cross_height = 4;
    for k = 1:length(Ell_Contour_IV)
        
        y = round(Ell_Para_IV(k).y);
        x = round(Ell_Para_IV(k).x);
        
        i_begin = max(1, y - cross_height);
        j_begin = max(1, x - cross_height);
        i_end = min(n_height, y + cross_height);
        j_end = min(n_width, x + cross_height);
        
        for i = i_begin:i_end
            for j = j_begin:j_end
                
                now_width = min(abs(i - y), abs(j - x));
                now_height = max(abs(i - y), abs(j - x));
                if (now_width <= cross_width) && (now_height <= cross_height)
                    Mark_image(i, j) = 1;
                end
            end
        end
    end
    
    Result_image = zeros(n_height, n_width, 3);
    Result_image(:, :, 1) = (Mark_image == 0) .* Raw_image + (Mark_image == 1) .* 255;
    Result_image(:, :, 2) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .*  round(255 .* (Mark_image - 2));
    Result_image(:, :, 3) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .* (128 + round(127 .* (3 - Mark_image)));
    Result_image = uint8(Result_image);
    imwrite(Result_image, [image_name, '_Fit_IV.jpg'], 'jpg');
end

% Step 5: Compare the two methods and output the reliable result
fprintf('5 Combining the two results via binary integer programming...\n\n');
Image_filtered_BF = double(Image_filtered);
[Ell_Para_Best, Ell_Contour_Best ,Ell_Rel_Best] = Evaluate_Result(Ell_Para_Cov, Ell_Contour_Cov, Ell_Para_IV, Ell_Contour_IV, Image_filtered_BF, dia_particle);
% Output result

Mark_image = zeros(n_height, n_width);
line_width = 1;
for k = 1:length(Ell_Contour_Best)
    boundary = Ell_Contour_Best{k};
    for l = 1 : length(boundary)
        
        y = boundary(l,1);
        x = boundary(l,2);
        
        i_begin = max(1, y - line_width);
        j_begin = max(1, x - line_width);
        i_end = min(n_height, y + line_width);
        j_end = min(n_width, x + line_width);
        for i = i_begin:i_end
            for j = j_begin:j_end
                
                now_width = sqrt((i - y)^2 + (j - x)^2);
                
                if now_width <= line_width
                    Mark_image(i, j) = 2 + Ell_Rel_Best(k);
                end
            end
        end
    end
end
cross_width = 1;
cross_height = 8;
for k = 1:length(Ell_Contour_Best)
    
    y = round(Ell_Para_Best(k).y);
    x = round(Ell_Para_Best(k).x);
    
    i_begin = max(1, y - cross_height);
    j_begin = max(1, x - cross_height);
    i_end = min(n_height, y + cross_height);
    j_end = min(n_width, x + cross_height);
    
    for i = i_begin:i_end
        for j = j_begin:j_end
            
            now_width = min(abs(i - y), abs(j - x));
            now_height = max(abs(i - y), abs(j - x));
            if (now_width <= cross_width) && (now_height <= cross_height)
                Mark_image(i, j) = 1;
            end
        end
    end
end

if b_intermediary == 1
    Result_image = zeros(n_height, n_width, 3);
    Result_image(:, :, 1) = (Mark_image == 0) .* Raw_image + (Mark_image == 1) .* 255 + (Mark_image > 3) .* 255;
    Result_image(:, :, 2) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2 & Mark_image <= 3) .*  round(255 .* (Mark_image - 2)) + (Mark_image > 3) .* 255;
    Result_image(:, :, 3) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2 & Mark_image <= 3) .* (128 + round(127 .* (3 - Mark_image)));
    Result_image = uint8(Result_image);
    imwrite(Result_image, [image_name, '_Fit_Best.jpg'], 'jpg');
end

Result_image = zeros(n_height, n_width, 3);
Result_image(:, :, 1) = (Mark_image == 0) .* Raw_image + (Mark_image == 1) .* 255;
Result_image(:, :, 2) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .* 255;
Result_image(:, :, 3) = (Mark_image == 0) .* Raw_image + (Mark_image >= 2) .* 128;
Result_image = uint8(Result_image);
imwrite(Result_image, [image_name, '_Result.jpg'], 'jpg');


% Output the total time
time_total = round(toc);
time_sec = mod(time_total, 60);
time_min = floor((time_total - time_sec) / 60);

fprintf('The total time of the nano-particle detection is: %dm%ds\n', time_min, time_sec);

